import numpy as np
import cv2


def filtration_box(img, filter_size):
    """
    Filters image noise using box blur algorithm

    Parameters
    ----------
    img : numpy.ndarray
        Input image.
    filter_size : int
        Size of box blur filter.
    Returns
    -------
    Output image.
    """
    return cv2.blur(img, (filter_size, filter_size))


def filtration_median(img, filter_size):
    """
    Filters image noise using median algorithm

    Parameters
    ----------
    img : numpy.ndarray
        Input image.
    filter_size : int
        Odd number - size of median filter.
    Returns
    -------
    Output image.
    """
    return cv2.medianBlur(img, filter_size)


def filtration_gauss(img, filter_size, sigma_x):
    """
    Filters image noise using Gaussian blur algorithm

    Parameters
    ----------
    img : numpy.ndarray
        Input image.
    filter_size : int
        Size of Gaussian filter.
    Returns
    -------
    Output image.
    """
    return cv2.GaussianBlur(img, (filter_size, filter_size), sigma_x)


def apply_fft(image):
    """
    Applies FFT on image given.

    Parameters
    ----------
    image : 2D array
        Image to perform FFT on.
    Returns
    -------
    mag_spec : 2D array
        Normalized magnitude spectrum.
    fftcls_shift : 2D array
        Centered product of FFT.
    """
    fftcls = np.fft.fft2(image)
    fftcls_shift = np.fft.fftshift(fftcls)
    mag_spec = 20 * np.log(np.abs(fftcls_shift))
    return cv2.normalize(mag_spec, None, 0, 255, cv2.NORM_MINMAX, cv2.CV_8U), fftcls_shift


def inverse_fft(fft_shift, filter_mask=None):
    """
    Applies inverse FFT.

    Parameters
    ----------
    fft_shift : 2D array
        Shifted computed FFT
    filter_mask : 2D array
        2D array mask containing 255 and 0 values.
    Returns
    -------
    img_back : 2D array
        Image made by inverse FFT.
    """
    fftshift = np.copy(fft_shift)
    if not filter_mask is None:
        fftshift[filter_mask != 255] = 0

    f_ishift = np.fft.ifftshift(fftshift)
    return np.abs(np.fft.ifft2(f_ishift))


def create_filter_mask(size, rows, columns):
    """
    Creates a filter mask specified by rows and columns. Specified rows and columns are set to 255, others 0.

    Parameters
    ----------
    size : tuple
        Size of resulting filter mask image.
    Returns
    -------
    filter_mask : 2D array
        2D array mask containing 255 and 0 values.
    """
    if type(size) != tuple:
        raise Exception('Size param must be tuple!')

    filter_mask = np.zeros(size, dtype=np.uint8)
    filter_mask[rows] = 255
    filter_mask[:, columns] = 255

    return filter_mask


def filter_mag_spec(mag_spec, filter_mask):
    """
    Filters input spectrum using filter_mask image.

    Parameters
    ----------
    mag_spec : 2D ndarray
        Image with magnitude spectrum.
    filter_mask : 2D array
        Filter binary mask image containing values to keep (255) and filter out (0).
    Returns
    -------
    result : 2D ndarray
        Vizualization of spectrum after filtering.
    """
    result = np.copy(mag_spec)
    result[filter_mask != 255] = 0

    return result



import numpy as np
import cv2

def apply_mask(img, mask_bin):
    """
    Masks colored image with binary mask. Output image is just logical AND between two images.
    """
    return cv2.bitwise_and(img, img, mask=mask_bin)


def logical_and(bin_im, bin_mask):
    return cv2.bitwise_and(bin_im, bin_mask)


def to_intensity(hue_angle):
    """
    Converts color angle in HUE definition into intensity value of brightness image in opencv.
    https://www.docs.opencv.org/trunk/df/d9d/tutorial_py_colorspaces.html

    Parameters
    ----------
    hue_angle : int
        Angle in HUE definition (0-359).
    Returns
    -------
    Integer value that represents the same HUE value but in opencv brightness image (0-179).
    """
    return int(hue_angle * 0.5)


def to_angle(hue_intensity):
    """
    Converts hue intensity value of brightness image in opencv into hue angle in HUE definition.
    https://www.docs.opencv.org/trunk/df/d9d/tutorial_py_colorspaces.html

    Parameters
    ----------
    hue_intensity : int
        Intensity value of brightness image (0-179).
    Returns
    -------
    Integer value that represents the HUE angle (0-359).
    """
    return hue_intensity * 2


def to_3_channels(image):
    """
    Converts 1 channel image to 3 channels.
    """
    if len(image.shape) == 3:
        raise Exception('Image already has 3 channels! Use it on binary or grayscale image only.')
    return cv2.merge([image, image, image])


def segmentation_one_threshold(img, threshold):
    """
    Segments image into black & white using one threshold

    Parameters
    ----------
    img : ndarray
        Input image.
    threshold : int
        Pixels with value lower than threshold are considered black, the others white.
    Returns
    -------
    Output image.
    """
    _, dst = cv2.threshold(img, threshold, 255, cv2.THRESH_BINARY)
    return dst


def segmentation_auto_threshold(img):
    """
    Segments image into black & white using automatic threshold

    Parameters
    ----------
    img : ndarray
        Input image.
    Returns
    -------
    Output image.
    """
    _, dst = cv2.threshold(img, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)
    return dst


def segmentation_two_thresholds(img, lower, higher):
    """
    Segments image into black & white using two thresholds

    Parameters
    ----------
    img : ndarray
        Input image.
    lower : int
        Pixels with value lower than threshold are considered black, the others white.
    higher : int
        Pixels with value higher than threshold are considered black, the others white.
    Returns
    -------
    Output image.
    """
    return cv2.inRange(img, min(lower, higher), max(lower, higher))


def segmentation_adaptive_threshold(img, size, constant=0):
    """
    Segments image into black & white using calculated adaptive
    threshold using Gaussian function in pixel neighbourhood.

    Parameters
    ----------
    img : ndarray
        Input image.
    size : int
        Size of used gaussian. Lowest value is 3. Algorithm uses only odd numbers.
    constant : int
        Value that is added to calculated threshlod. It could be negative as well as zero as well as positive number.
    Returns
    -------
    img : ndarray
        Output binary image.
    """
    if size < 3:
        size = 3
    elif size % 2 == 0:
        size -= 1
    return cv2.adaptiveThreshold(img, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY, size, int(constant))



from qreader import QReader
import cv2


def __GRAY2BGR(img):
    """ 
    If the image has shape less than 3, tries to convert it to BGR. Otherwise returns the image unchanged.

    Parameters
    ----------
    img : np.ndarray
        Input image to be converted to BGR.
        
    Returns
    -------
    img: np.ndarray
        Output image in BGR format or in the original format if the conversion failed of the shape was already 3.
    conversion_took_place: bool
        True if the conversion took place, False otherwise.
    """    
    if len(img.shape) < 3:
        try:
            img_bgr = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
            
            return img_bgr, True
        except:
            pass
    
    return img, False
    

def qr_init_reader(model_size = 's', min_confidence = 0.5, reencode_to = 'shift_jis'):
    """
    Initializes a QR code reader. If you want to use the detect or decode QR codes on multiple occasions, it is recommended to initialize the reader once and then pass it to the functions.
    For further info refer to: https://github.com/Eric-Canas/qreader
    
    Parameters
    ----------
    model_size : str
        The size of the model to use. It can be 'n' (nano), 's' (small), 'm' (medium) or 'l' (large). Larger models are more accurate but slower. Defaults to 's'.
    min_confidence : float 
        The minimum confidence of the QR detection to be considered valid. Values closer to 0.0 can get more False Positives, while values closer to 1.0 can lose difficult QRs. Default (and recommended): 0.5.
    reencode_to : str | None 
        The encoding to reencode the utf-8 decoded QR string. 
        If None, it won't re-encode. 
        If you find some characters being decoded incorrectly, try to set a Code Page (https://learn.microsoft.com/en-us/windows/win32/intl/code-page-identifiers) that matches your specific charset. 
        Recommendations that have been found useful:
            'shift-jis' for Germanic languages
            'cp65001' for Asian languages  Defaults to 'shift_jis'.

    Returns
    -------
    reader : QReader
        Initialized QR code reader.
    """  
    return QReader(model_size = model_size, min_confidence = min_confidence, reencode_to = reencode_to)
    

def qr_detect_and_decode(img, return_detections = False, is_bgr = True, reader = qr_init_reader()):
    """
    Detects and decodes QR codes in the given image and return the decoded strings (or None, if any of them was detected but not decoded).
    For further info refer to: https://github.com/Eric-Canas/qreader
    
    Parameters
    ----------
    img : np.ndarray
        The image to be read. It is expected to be RGB or BGR (uint8). Format (HxWx3). Can also be grayscale (HxW), in which case it will be converted to BGR.
    return_detections : bool
        If True, it will return the full detection results together with the decoded QRs. If False, it will return only the decoded content of the QR codes. Defaults to False.
    is_bgr : bool
        If True, the received image is expected to be BGR instead of RGB. Defaults to True.
    reader : QReader
        Initialized QReader class, use qr_init_reader(model_size, min_confidence, reencode_to) with different parameters if you wish. Defaults to qr_init_reader().

    Returns
    -------
    detections : tuple[dict[str, np.ndarray | float | tuple[float | int, float | int]], str | None] OR decoded_strings : tuple[str | None, ...]
        If return_detections is True, it will return a tuple of tuples. Each tuple will contain the detection result (a dictionary with the keys 'confidence', 'bbox_xyxy', 'polygon_xy'...) and the decoded QR code (or None if it can not be decoded). 
        If return_detections is False, it will return a tuple of strings with the decoded QR codes (or None if it can not be decoded).
    """    
    # if the image is in grayscale, convert it to BGR
    img, converted  = __GRAY2BGR(img)  
    return reader.detect_and_decode(img, return_detections = return_detections, is_bgr = is_bgr if converted == False else True)

def qr_decode(img, detection_result, is_bgr = True, reader = qr_init_reader()):
                
    """
    Decodes a single QR code on the given image, described by a detection_result.
    For further info refer to: https://github.com/Eric-Canas/qreader
    
    Internally, this method will run the pyzbar decoder, using the information of the detection_result, to apply
    different image preprocessing techniques that heavily increase the decoding rate.

    Parameters
    ----------
    image : np.ndarray
        The image to be read. It is expected to be RGB or BGR (uint8). Format (HxWx3). Can also be grayscale (HxW), in which case it will be converted to BGR.
    detection_result : dict[str, np.ndarray|float|tuple[float|int, float|int]]
        One of the detection dicts returned by the detect method. Note that qr_detect() returns a tuple of these dicts. This method expects just one of them.
    is_bgr : bool 
        If True, the received image is expected to be BGR instead of RGB. Defaults to True.
    reader : QReader
        Initialized QReader class, use qr_init_reader(model_size, min_confidence, reencode_to) with different parameters if you wish. Defaults to qr_init_reader().
    
    Returns
    -------
    decoded_strings : str OR None 
        The decoded content of the QR code or None if it can not be read.
    """    
    # if the image is in grayscale, convert it to BGR
    img, converted = __GRAY2BGR(img)  
        
    if is_bgr or converted == True:
        # decode expects RGB images
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    return reader.decode(img, detection_result)

def qr_detect(img, is_bgr = True, reader = qr_init_reader()):
    """
    Detects QR codes in the image and returns a tuple of dictionaries with all the detection information.
    For further info refer to: https://github.com/Eric-Canas/qreader

    Parameters
    ----------
    image : np.ndarray 
        The image to be read. It is expected to be RGB or BGR (uint8). Format (HxWx3).
    is_bgr : bool 
        If True, the received image is expected to be BGR instead of RGB. Defaults to True.
    reader : QReader 
        Initialized QReader class, use qr_init_reader(model_size, min_confidence, reencode_to) with different parameters if you wish. Defaults to qr_init_reader().
        
    Returns
    -------
    detections : tuple[dict[str, np.ndarray|float|tuple[float|int, float|int]]]: 
        A tuple of dictionaries containing the following keys:
            - 'confidence': float. The confidence of the detection.
            - 'bbox_xyxy': np.ndarray. The bounding box of the detection in the format [x1, y1, x2, y2].
            - 'cxcy': tuple[float, float]. The center of the bounding box in the format (x, y).
            - 'wh': tuple[float, float]. The width and height of the bounding box in the format (w, h).
            - 'polygon_xy': np.ndarray. The accurate polygon that surrounds the QR code, with shape (N, 2).
            - 'quad_xy': np.ndarray. The quadrilateral that surrounds the QR code, with shape (4, 2). Fitted from the polygon.
            - 'padded_quad_xy': np.ndarray. An expanded version of quad_xy, with shape (4, 2), that always include all the points within polygon_xy.

            All these keys (except 'confidence') have a 'n' (normalized) version. For example, 'bbox_xyxy' is the bounding box in absolute coordinates, while 'bbox_xyxyn' is the bounding box in normalized coordinates (from 0. to 1.).
    """    
    # if the image is in grayscale, convert it to BGR
    img, converted = __GRAY2BGR(img)  
    return reader.detect(img, is_bgr = is_bgr if converted == False else True)







import PIL
from pytesseract import pytesseract

from improutils import negative


def ocr(img_bin, config=''):
    """
    Detects text in the file.

    Parameters
    ----------
    img_bin : ndarray
        Input binary image. White objects on black background.
    config : str
        Model config, refer to: https://tesseract-ocr.github.io/tessdoc/ImproveQuality.html, https://muthu.co/all-tesseract-ocr-options/ for correct use.  Defaults to ''.
    Returns
    -------
    The recognized text in the image.
    """
    # Tesseract works with black objects on white background.
    img_bin = negative(img_bin)
    return pytesseract.image_to_string(PIL.Image.fromarray(img_bin), config=config)



import numpy as np

# Functions for conversion within different types of coordinates
def convert_pt_to_homogenous(pt):
    """
    Converts the input point from inhomogeneous coordinates to homogeneous ones.

    Parameters
    ----------
    pt : ndarray
        Input point in inhomogeneous coordinates.

    Returns
    -------
    _ : ndarray
        Input point in homogeneous coordinates.
    """
    return np.append(pt, np.array(1))

def convert_pt_from_homogenous(pt):
    """
    Convert input point in homogeneous coordinates to inhomogeneous.

    Parameters
    ----------
    pt : ndarray
        Input point in homogeneous coordinates.

    Returns
    -------
    _ : tuple
        Input point in inhomogeneous coordinates.
    """
    return tuple([elem / pt[-1] for elem in pt[:-1]])

def convert_pts_to_homogenous(pts):
    """
    Convert input points in inhomogeneous coordinates to homogeneous.

    :param pt: ndarray
        Input points in inhomogeneous coordinates.
    :return: ndarray
        Input points in homogeneous coordinates.
    """
    return np.array([convert_pt_to_homogenous(pt) for pt in pts])

def convert_pts_from_homogenous(pts):
    """
    Convert input points in homogeneous coordinates to inhomogeneous.

    :param pt: ndarray
        Input points in homogeneous coordinates.
    :return: ndarray
        Input points in inhomogeneous coordinates.
    """
    return np.array([convert_pt_from_homogenous(pt) for pt in pts])

def _calc_alfa_metric_factor(ref_measurements, vanish_line, vert_vanish_point):
    """
    Calculates alfa metric factor using multiple reference measurements via minimization ||As|| = 0. This is done by SVD.
        In depth overview can be found in https://www.robots.ox.ac.uk/~vgg/publications/1999/Criminisi99b/criminisi99b.pdf - PDF page 104.

    :param ref_measurements: list
        Each measurement is in (t_ref, b_ref, height) format. ``Image coordinates are in inhomogeneous format.
    :param vanish_line: ndarray
        Homogenous coordinates of vanishing line.
    :param vert_vanish_point: ndarray
        Homogenous coordinates of vanishing point in reference direction.
    :return: float
        Scalar value of alfa metric factor calculated by SVD.
    """
    matrix_A = np.empty((len(ref_measurements), 2), dtype='float64')

    for i, (t_ref, b_ref, h_ref) in enumerate(ref_measurements):
        t_ref = convert_pt_to_homogenous(t_ref)
        b_ref = convert_pt_to_homogenous(b_ref)
        beta = np.linalg.norm(np.cross(b_ref, t_ref))
        ro = np.dot(vanish_line, b_ref)
        gamma = np.linalg.norm(np.cross(vert_vanish_point, t_ref))
        matrix_A[i] = (h_ref * ro * gamma, beta)
        # alfa_metric_factor = - np.linalg.norm(np.cross(b_ref, t_ref)) / \
        #                               (h_ref * (np.dot(vanish_line, b_ref)) * np.linalg.norm(np.cross(vert_vanish_point, t_ref)))
        # print(alfa_metric_factor)

    u, s, vh = np.linalg.svd(matrix_A)
    return vh[0, -1] / vh[1, -1]



import numpy as np
import cv2

def contour_to_image(contour, image, size=None):
    """
    Creates a new image from the contour.
    It's similar to contour cropping but it's not that fast.
    It does not suffer from the known error if the contour is irregulary shaped.

    Parameters
    ----------
    contour : ndarray
        Contour that represents the area from image to be cropped.
    img_bin : ndarray
        Input binary image.
    size : tuple
        Optional size of the created image.
        If it's not used, the image's size is the same as the
        size of bounding rectangle of the input contour.
    Returns
    -------
    Output cropped image.
    """

    if size is None:
        _, _, w, h = cv2.boundingRect(contour)
        size = (w, h)

    assert type(size) is tuple, 'Param size should be a tuple!'
    blank = np.zeros_like(image)
    half_x = int(size[0] * 0.5)
    half_y = int(size[1] * 0.5)

    c = get_center(contour)
    cv2.drawContours(blank, [contour], -1, (255, 255, 255), cv2.FILLED)

    return blank[c[1] - half_y:c[1] + half_y, c[0] - half_x:c[0] + half_x].copy()


def find_contours(img_bin, min_area=0, max_area=np.inf, fill=True, external=True):
    """
    Finds contours in binary image and filters them using their area. Then it draws binary image
    from filtered contours. It counts contours as well.

    Parameters
    ----------
    img_bin : ndarray
        Input binary image.
    min_area : int
        Size of contour that is used to filter all smaller contours out.
    max_area : int
        Size of contour that is used to filter all larger contours out.
    Returns
    -------
    contour_drawn : ndarray
        Output binary image with drawn filled filtered contours.
    count : int
        Number of found and filtered contours.
    contours : list
        Found contours.
    """
    mode = cv2.RETR_EXTERNAL
    if not external:
        mode = cv2.RETR_LIST
    contours, _ = cv2.findContours(img_bin, mode, cv2.CHAIN_APPROX_SIMPLE)
    contours = [c for c in contours if cv2.contourArea(c) > min_area and cv2.contourArea(c) < max_area]
    thick = cv2.FILLED
    if not fill: thick = 2
    contour_drawn = cv2.drawContours(np.zeros(img_bin.shape, dtype=np.uint8), contours, -1, color=(255, 255, 255),
                                     thickness=thick)
    return contour_drawn, len(contours), contours


def fill_holes(img_bin, close=False, size=5):
    """
    Fill the holes in found contours. It could merge the contour using close input with appropriate size.

    Parameters
    ----------
    img_bin : ndarray
        Input binary image.
    close : boolean
        If it should merge contours with missing points using close operation.
    size : int
        Size of the close operation element.
    Returns
    -------
    Output binary image.
    """
    if close:
        struct = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (size, size))
        img_bin = cv2.morphologyEx(img_bin, cv2.MORPH_CLOSE, struct)
    res, _, _ = find_contours(img_bin)
    return res


def get_center(contour):
    """
    Gets the center of a contour in pixels in tuple format.

    Parameters
    ----------
    contour : ndarray
        input contour.
    Returns
    -------
    _ : tuple
        Center in pixels in tuple format.
    """
    M = cv2.moments(contour)
    cX = int(M['m10'] / M['m00'])
    cY = int(M['m01'] / M['m00'])

    return cX, cY



import numpy as np
import cv2

# Dimensionless descriptors
from improutils import find_contours


class ShapeDescriptors:
    """
    An internal class for shape descriptors.
    If you are an user, you shall not call not nor work with the class directly
    """
    def form_factor(area, perimeter):
        return (4 * np.pi * area) / (perimeter * perimeter)

    def roundness(area, max_diameter):
        return (4 * area) / (np.pi * max_diameter * max_diameter)

    def aspect_ratio(min_diameter, max_diameter):
        return min_diameter / max_diameter;

    def convexity(perimeter, convex_perimeter):
        return convex_perimeter / perimeter

    def solidity(area, convex_area):
        return area / convex_area

    def compactness(area, max_diameter):
        return np.sqrt(4 / np.pi * area) / max_diameter;

    def extent(area, bounding_rectangle_area):
        return area / bounding_rectangle_area;

"""
An internal, helper function.
Shall not be called directly by the user.
This can perform a check for contour validity.

Right now, there is no check for validity, and the function does nothing.

If the contour would be invalid, a ValueError shall be risen

Parameters
----------
contour : ndarray
    One contour.

Returns
-------

Throws
-------
_ : a ValueError exception if ???
"""
def _validateContourGiven(contour):
    return;

"""
Aka Špičatost.
Allows to determine the form factor of the contour given
Parameters
----------
contour : ndarray
    The contour. You can get it by calling
    _, _, contours = improutils.find_contours()
    Then that one contours is contours[i] where i is index of your choice.

Returns
-------
_ : number
    The number, describing the contour property

"""
def form_factor(contour):
    _validateContourGiven(contour)
    return ShapeDescriptors.form_factor(cv2.contourArea(contour), cv2.arcLength(contour, True))

"""
Allows to determine the roundness of the contour given
Parameters
----------
contour : ndarray
    The contour. You can get it by calling
    _, _, contours = improutils.find_contours()
    Then that one contours is contours[i] where i is index of your choice.

Returns
-------
_ : number
    The number, describing the contour property

"""
def roundness(contour):
    _validateContourGiven(contour)
    area = cv2.contourArea(contour)
    _, radius = cv2.minEnclosingCircle(contour)
    r = ShapeDescriptors.roundness(area, 2 * radius)
    if r > 1:
        r = 1
    return r

"""
Allows to determine the aspect ratio of the contour given
Parameters
----------
contour : ndarray
    The contour. You can get it by calling
    _, _, contours = improutils.find_contours()
    Then that one contours is contours[i] where i is index of your choice.

Returns
-------
_ : number
    The number, describing the contour property

"""
def aspect_ratio(contour):
    _validateContourGiven(contour)
    dims = cv2.minAreaRect(contour)[1]
    min_diameter = min(dims)
    max_diameter = max(dims)
    return ShapeDescriptors.aspect_ratio(min_diameter, max_diameter)


"""
Allows to determine the convexity of the contour given
Parameters
----------
contour : ndarray
    The contour. You can get it by calling
    _, _, contours = improutils.find_contours()
    Then that one contours is contours[i] where i is index of your choice.

Returns
-------
_ : number
    The number, describing the contour property

"""
def convexity(contour):
    _validateContourGiven(contour)
    hull = cv2.convexHull(contour, None, True, True)
    per = cv2.arcLength(contour, True)
    conv_per = cv2.arcLength(hull, True)
    r = ShapeDescriptors.convexity(per, conv_per)
    if r > 1:
        r = 1
    return r


"""
Allows to determine the solidity of the contour given
Parameters
----------
contour : ndarray
    The contour. You can get it by calling
    _, _, contours = improutils.find_contours()
    Then that one contours is contours[i] where i is index of your choice.

Returns
-------
_ : number
    The number, describing the contour property

"""
def solidity(contour):
    _validateContourGiven(contour)
    hull = cv2.convexHull(contour, None, True, True)
    area = cv2.contourArea(contour)
    conv_area = cv2.contourArea(hull)
    r = ShapeDescriptors.solidity(area, conv_area)
    if r > 1: r = 1
    return r


"""
Allows to determine the compactness of the contour given
Parameters
----------
contour : ndarray
    The contour. You can get it by calling
    _, _, contours = improutils.find_contours()
    Then that one contours is contours[i] where i is index of your choice.

Returns
-------
_ : number
    The number, describing the contour property

"""
def compactness(contour):
    _validateContourGiven(contour)
    area = cv2.contourArea(contour)
    max_diameter = max(cv2.minAreaRect(contour)[1])
    r = ShapeDescriptors.compactness(area, max_diameter)
    if r > 1: r = 1
    return r


"""
Allows to determine the extent of the contour given
Parameters
----------
contour : ndarray
    binary image. This image contains only black and white values.
    Traditionally, you get it from the segmentation process.

Returns
-------
_ : number
    The number, describing the contour property

"""
def extent(contour):
    _validateContourGiven(contour)
    area = cv2.contourArea(contour)
    w, h = cv2.minAreaRect(contour)[1]
    return ShapeDescriptors.extent(area, w * h)



import os
import cv2
import numpy as np
import yaml

from improutils.visualisation import plot_images

# consts
IDX_CAM_MATRIX = "camera_matrix"
IDX_DIST_COEFFS = "dist_coefs"

def camera_calib(input_source, chess_shape, output_calib_file=None, img_show_delay=1):
    """
    Browses all images found in input_source and on each image tries to find chessboard corners.
    If chessboard corners are found, image corespondences with real world space are added to lists.
    Based on these image-world corespondences, camera calibration is made.

    Parameters
    ----------
    input_source : string
        Input source for cv2.VideoCapture (could be camera source or sequence of saved images where format has to be specified).
    chess_shape : tuple
        Number of inner corners per a chessboard row and column.
    output_calib_file : Optional[string]
        Output file where calibration is saved when neccesary.
    img_show_delay : int
        Delay in ms between shown images.
    Returns
    -------
    _ : tuple
        camera matrix and distance coefficients
    """

    cap = cv2.VideoCapture(input_source)

    if not cap.isOpened():
        cv2.destroyAllWindows()
        raise FileNotFoundError('Capture cannot be opened.')

    # prepare object points, like (0,0,0), (1,0,0), (2,0,0) ....,(6,5,0)
    objp = np.zeros((chess_shape[0] * chess_shape[1], 3), np.float32)
    objp[:, :2] = np.mgrid[0:chess_shape[0], 0:chess_shape[1]].T.reshape(-1, 2)

    # Arrays to store object points and image points from all the images.
    objpoints = []  # 3d point in real world space
    imgpoints = []  # 2d points in image plane.

    cntr = 0
    good_images = []

    while cap.isOpened():
        print(f"Img {cntr} is being processed..")
        cntr += 1
        ret, frame = cap.read()
        if not ret: break

        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

        # If found, add object points, image points (after refining them)
        ret, corners = cv2.findChessboardCorners(gray, chess_shape)

        if ret:
            objpoints.append(objp)
            corners = cv2.cornerSubPix(gray, corners, (11, 11), (-1, -1),
                                       (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.1))
            imgpoints.append(corners)

            # Draw and display the corners
            frame = cv2.drawChessboardCorners(frame, chess_shape, corners, ret)
            good_images.append(frame)

    print(f'{len(good_images)} from {cntr} frames were correctly detected.')
    plot_images(good_images[0])
    print('Computing camera matrix...')
    rms, camera_matrix, dist_coefs, rvecs, tvecs = cv2.calibrateCamera(objpoints, imgpoints, gray.shape[::-1], None,
                                                                       None)

    if output_calib_file is not None:
        save_camera_calib(output_calib_file, camera_matrix, dist_coefs)

    print('\nRMS:', rms)
    print('Camera matrix:\n', camera_matrix)
    print('Distortion coefficients: ', dist_coefs.ravel())

    return camera_matrix, dist_coefs, good_images


def correct_frame(frame, camera_matrix, dist_coeffs):
    """
    Returns an undistorted frame.
    """
    return cv2.undistort(frame, camera_matrix, dist_coeffs)

def load_camera_calib(input_file):
    """
    Loads camera calibration from specified input file.

    Parameters
    ----------
    input_file : string
        Input file with calibration data in YAML format.
    Returns
    -------
    tuple(ndarray, ndarray)
        Returns a tuple where first element is camera matrix array and second element is dist coefficients array.
        These arrays might be empty if the file isn't found or in correct format.
    """
    try:
        with open(input_file, 'r') as stream:
            data = yaml.load(stream)
            return data[IDX_CAM_MATRIX], data[IDX_DIST_COEFFS]
    except (FileNotFoundError, yaml.YAMLError) as exc:
        print(f'File {input_file} couldn\'t be read.')
        return np.array([]), np.array([])


def save_camera_calib(output_file, camera_matrix, dist_coefs):
    """
    Saves camera calibration to specified output file.

    Parameters
    ----------
    output_file : string
        Output file used for storing calibration data in YAML format. Parent directory is created if needed.
    Returns
    -------
    None
    """
    data = {IDX_CAM_MATRIX: camera_matrix, IDX_DIST_COEFFS: dist_coefs}
    output_dir = os.path.dirname(output_file)

    if not os.path.isdir(output_dir):
        os.mkdir(output_dir)

    with open(output_file, "w") as f:
        yaml.dump(data, f)



# Configuration PyPylon Viewer to load features for RGB Matrix Camera
VIEWER_CONFIG_RGB_MATRIX = {
    "features": [
        {
            "name": "GainRaw",
            "type": "int",
            "step": 1,
        },
        {
            "name": "Height",
            "type": "int",
            "unit": "px",
            "step": 2,
        },
        {
            "name": "Width",
            "type": "int",
            "unit": "px",
            "step": 2,
        },
        {
            "name": "CenterX",
            "type": "bool",
        },
        {
            "name": "CenterY",
            "type": "bool",

        },
        {
            "name": "OffsetX",
            "type": "int",
            "dependency": {"CenterX": False},
            "unit": "px",
            "step": 2,
        },
        {
            "name": "OffsetY",
            "type": "int",
            "dependency": {"CenterY": False},
            "unit": "px",
            "step": 2,
        },
        {
            "name": "AcquisitionFrameRateAbs",
            "type": "int",
            "unit": "fps",
            "dependency": {"AcquisitionFrameRateEnable": True},
            "value": 30,
            "max": 150,
            "min": 1,
        },
        {
            "name": "AcquisitionFrameRateEnable",
            "type": "bool",
        },
        {
            "name": "ExposureAuto",
            "type": "choice_text",
            "options": ["Off", "Once", "Continuous"],
            "style": {"button_width": "90px"}
        },
        {
            "name": "ExposureTimeAbs",
            "type": "int",
            "dependency": {"ExposureAuto": "Off"},
            "unit": "μs",
            "step": 100,
            "max": 35000,
            "min": 500,
        },
        {
            "name": "BalanceWhiteAuto",
            "type": "choice_text",
            "options": ["Off", "Once", "Continuous"],
            "style": {"button_width": "90px"}
        },
    ],
    "features_layout": [
        ("Height", "Width"),
        ("OffsetX", "CenterX"),
        ("OffsetY", "CenterY"),
        ("ExposureAuto", "ExposureTimeAbs"),
        ("AcquisitionFrameRateAbs", "AcquisitionFrameRateEnable"),
        ("BalanceWhiteAuto", "GainRaw")
    ],
    "actions_layout": [
        ("StatusLabel"),
        ("SaveConfig", "LoadConfig", "ContinuousShot", "SingleShot"),
        ("UserSet")
    ],
    "default_user_set": "UserSet3",
}

# Configuration PyPylon Viewer to load features for Monochromatic Matrix Camera
VIEWER_CONFIG_MONO_MATRIX = {
    "features": [
        {
            "name": "GainRaw",
            "type": "int",
            "step": 1,
        },
        {
            "name": "Height",
            "type": "int",
            "unit": "px",
            "step": 2,
        },
        {
            "name": "Width",
            "type": "int",
            "unit": "px",
            "step": 2,
        },
        {
            "name": "CenterX",
            "type": "bool",
        },
        {
            "name": "CenterY",
            "type": "bool",

        },
        {
            "name": "OffsetX",
            "type": "int",
            "dependency": {"CenterX": False},
            "unit": "px",
            "step": 2,
        },
        {
            "name": "OffsetY",
            "type": "int",
            "dependency": {"CenterY": False},
            "unit": "px",
            "step": 2,
        },
        {
            "name": "AcquisitionFrameRateAbs",
            "type": "int",
            "unit": "fps",
            "dependency": {"AcquisitionFrameRateEnable": True},
            "value": 30,
            "max": 150,
            "min": 1,
        },
        {
            "name": "AcquisitionFrameRateEnable",
            "type": "bool",
        },
        {
            "name": "ExposureAuto",
            "type": "choice_text",
            "options": ["Off", "Once", "Continuous"],
            "style": {"button_width": "90px"}
        },
        {
            "name": "ExposureTimeAbs",
            "type": "int",
            "dependency": {"ExposureAuto": "Off"},
            "unit": "μs",
            "step": 500,
            "max": 35000,
            "min": 500,
        },
    ],
    "features_layout": [
        ("Height", "Width"),
        ("OffsetX", "CenterX"),
        ("OffsetY", "CenterY"),
        ("ExposureTimeAbs", "ExposureAuto"),
        ("AcquisitionFrameRateAbs", "AcquisitionFrameRateEnable"),
        ("GainRaw")
    ],
    "actions_layout": [
        ("StatusLabel"),
        ("SaveConfig", "LoadConfig", "ContinuousShot", "SingleShot"),
        ("UserSet")
    ],
    "default_user_set": "UserSet3",
}

# Configuration PyPylon Viewer to load features for Monochromatic Line Scan Camera
VIEWER_CONFIG_MONO_LINE = {
    "features": [
        {
            "name": "GainRaw",
            "type": "int",
            "step": 1,
        },
        {
            "name": "Height",
            "type": "int",
            "unit": "px",
            "step": 2,
        },
        {
            "name": "Width",
            "type": "int",
            "unit": "px",
            "step": 2,
        },
        {
            "name": "CenterX",
            "type": "bool",
        },
        {
            "name": "OffsetX",
            "type": "int",
            "dependency": {"CenterX": False},
            "unit": "px",
            "step": 2,
        },
        {
            "name": "AcquisitionLineRateAbs",
            "type": "int",
            "max": 5000,
            "min": 100,
            "step": 100,
        },
        {
            "name": "ExposureAuto",
            "type": "choice_text",
            "options": ["Off", "Once", "Continuous"],
            "style": {"button_width": "90px"}
        },
        {
            "name": "ExposureTimeAbs",
            "type": "int",
            "dependency": {"ExposureAuto": "Off"},
            "unit": "μs",
            "step": 500,
            "max": 35000,
            "min": 500,
        },
    ],
    "features_layout": [
        ("Height", "Width"),
        ("OffsetX", "CenterX"),
        ("ExposureTimeAbs", "ExposureAuto"),
        ("AcquisitionLineRateAbs", "GainRaw")
    ],
    "actions_layout": [
        ("StatusLabel"),
        ("SaveConfig", "LoadConfig", "ContinuousShot", "SingleShot"),
        ("UserSet")
    ],
    "default_user_set": "UserSet3",
}

# Configuration PyPylon Viewer to load features for Monochromatic Matrix Camera with pericentric lens
VIEWER_CONFIG_MONO_MATRIX_PERICENTRIC = {
    "features": [
        {
            "name": "Gain",
            "type": "int",
            "unit": "dB",
            "step": 1,
        },
        {
            "name": "Height",
            "type": "int",
            "unit": "px",
            "step": 2,
            "value": 1270
        },
        {
            "name": "Width",
            "type": "int",
            "unit": "px",
            "step": 2,
            "value": 1344
        },
        {
            "name": "OffsetX",
            "type": "int",
            "unit": "px",
            "step": 2,
            "value": 608
        },
        {
            "name": "OffsetY",
            "type": "int",
            "unit": "px",
            "step": 2,
            "value": 418
        },
        {
            "name": "ExposureTime",
            "type": "int",
            "max": 35000,
            "min": 90,
            "step": 30,
            "value": 1500,
            "unit": "μs",
        },
    ],
    "features_layout": [
        ("Height", "Width"),
        ("OffsetX", "OffsetY"),
        ("Gain", "ExposureTime"),
    ],
    "actions_layout": [
        ("StatusLabel"),
        ("SaveConfig", "LoadConfig", "ContinuousShot", "SingleShot"),
        ("UserSet")
    ],
    "default_user_set": "UserSet3",
}



import numpy as np
import cv2
import os

def midpoint(ptA, ptB):
    """
    Returns the midpoint between two input 2D points.

    Parameters
    ----------
    ptA : array | tuple | ndarray
        The first 2D point considered
    ptA : array | tuple | ndarray
        The second 2D point considered

    Returns
    -------
    _ : tuple
        The 2D midpoint

    Throws
    -------
    _ : a ValueError exception if
        [+] Any of the given input points are not 2D.
        That means, the length of the structure (array|tuple|ndarray)
        is not equal to 2
    """

    if(len(ptA) != 2 or len(ptB) != 2):
        raise ValueError("Ivalid input point format");
    return ((ptA[0] + ptB[0]) * 0.5, (ptA[1] + ptB[1]) * 0.5)

def artificial_circle_image(size):
    """
    Creates an image with circles
    Parameters
    ----------
    size : int
        size of the image

    Returns
    -------
    _ : ndarray
        artificial image with circles
    Throws
    -------
    _ : a ValueError exception if any of the following contitions hold:
        [+] The input size is not an integer type
        [+] The input size is smaller than 1 (< 1)
    """
    if isinstance(size, int):
        raise ValueError("Ivalid input size. Integer required. got {}".format(type(size)));
    if size < 1:
        raise ValueError("Ivalid input size, must be >= 0. Got {}".format(size));

    img_art_circ = np.zeros((size, size), dtype=np.uint8)
    step = 10
    for i in range(step, size, step):
        cv2.circle(img_art_circ, (int(size / 2.0), int(size / 2.0)), i - step, np.random.randint(0, 255), thickness=4)
    return img_art_circ


def order_points(pts):
    """
    Sorts the points based on their coordinates,
    in top-left, top-right, bottom-right, and bottom-left order

    Parameters
    ----------
    pts : ndarray
        2D Points to be sorted.
        The points are expected to be in 2D cartesian plane coordinates.
        Must be type of ndarray.
        Must have length of 4 or more.

        Each element of the array is expected to be an ARRAY,
        containing exactly 2 elements, specifying the x and y, respectivelly

        If the content of the ndarray provided is not in the expected format,
        the behavior is not defined.

    Returns
    -------
    _ : ndarray
        sorted points, the coordinates in top-left, top-right, bottom-right, and bottom-left order
    Throws
    -------
    _ : a ValueError exception if any of the following contitions hold:
        [+] The input is not of an ndarray type
        [+] The input ndarray does have length below 4 (< 4)
        [+] The element (point) in the input ndarray does not have the dimension exactly of 2
    """

    if(not isinstance(pts, np.ndarray)):
        raise ValueError("Ivalid input point format. Numpy ndarray expected. Got {}".format(type(pts)));

    if(len(pts) < 4):
        raise ValueError("Ivalid amount of input points. Got {} elements".format(len(pts)));

    #understanding, that creating an NP array where not all elements have the same length
    #prints a warning thanks to python by itself,
    #we just check for the first element, and assume that they all have the same length
    if(len(pts[0]) != 2):
        raise ValueError("Ivalid input point format. For a point, dimension of 2 expected. got {}".format(len(pts[0])));


    xSorted = pts[np.argsort(pts[:, 0]), :]

    # grab the left-most and right-most points from the sorted
    # x-roodinate points
    leftMost = xSorted[:2, :]
    rightMost = xSorted[2:, :]

    # now, sort the left-most coordinates according to their
    # y-coordinates so we can grab the top-left and bottom-left
    # points, respectively
    leftMost = leftMost[np.argsort(leftMost[:, 1]), :]
    (bl, tl) = leftMost

    # now that we have the top-left coordinate, use it as an
    # anchor to calculate the Euclidean distance between the
    # top-left and right-most points; by the Pythagorean
    # theorem, the point with the largest distance will be
    # our bottom-right point
    rightMost = rightMost[np.argsort(rightMost[:, 1]), :]
    (br, tr) = rightMost

    # return the coordinates in top-left, top-right,
    # bottom-right, and bottom-left order
    return np.array([tl, tr, br, bl], dtype="float32")


def pcd_to_depth(pcd, height, width):
    """
    Reduce point-cloud to coordinates, point cloud [x, y, z, rgb] -> depth[x, y, z]

    Parameters
    ----------
    pcd : array
        point cloud
    height : int
        height of captured img
    width : int
        width of a captured img
    Returns
    ----------
    _ : array
        coordinates
    """
    data = pcd
    data = [float(x.split(' ')[2]) for x in data]
    data = np.reshape(data, (height, width))
    return data

def create_file_path(folder, file_name):
    """
    Easier defined function to create path for filename inside a folder.
    Parameters
    ----------
    folder : string
        Base folder directory in string notation.
        If the directory does not exist, it is created.

    file_name : string
        File name that should be inside the base folder.
    Returns
    -------
    string
        Path to the file.
    """
    if not os.path.isdir(folder):
        os.mkdir(folder)

    return os.path.join(folder, file_name)



import os

import numpy as np
import cv2

import re
from natsort import natsorted
import shutil


def load_image(file_path):
    """
    Loads an image from a file.

    The function calls cv2.imread() to load image from the specified file
    and then return it. If the image cannot be read, the AssertError exception is thrown.

    For more info about formats, see cv2.imread() documentation

    Parameters
    ----------
    file_path : string
        A path to an image file
    Returns
    -------
    Loaded image in numpy.ndarray
    """
    assert os.path.exists(file_path), 'File does NOT exist! (' + file_path + ')'
    return cv2.imread(file_path)


def save_image(image, file_path):
    """
    Save an image to a file.

    The function calls cv2.imwrite() to save an image to the specified file. The image format is chosen based on the
    filename extension.

    Parameters
    ----------
    image : numpy.ndarray
        Pixel values
    file_path : string
        A path to an image file
    Returns
    -------
    True if image is saved successfully
    """
    return cv2.imwrite(file_path, image)


def copy_to(src, dst, mask):
    """
    Copies source image pixel to destination image using mask matrix.

    This function is Python alternative to C++/Java OpenCV's Mat.copyTo().
    More: https://docs.opencv.org/trunk/d3/d63/classcv_1_1Mat.html#a626fe5f96d02525e2604d2ad46dd574f

    Parameters
    ----------
    src : numpy.ndarray
        Source image
    dst : numpy.ndarray
        Destination image
    mask : numpy.ndarray
        Binary image that specifies which pixels are copied. Value 1 means true
    Returns
    -------
    Destination image with copied pixels from source image
    """
    locs = np.where(mask != 0)  # Get the non-zero mask locations
    dst[locs[0], locs[1]] = src[locs[0], locs[1]]
    return dst

def reindex_image_files(source_dir, output_dir=None):
    """ Reads all images in source_dir and based on they original order,
    change their filename to be continuous integer (starting from 0).
    Then, they can be easily read by cv2.VideoCapture. Image format is kept.

    Parameters
    ----------
    source_dir : string
        Input images directory that have to be renamed.
    output_dir : Optional[string]
        Output directory for renamed files. If not specified, renaming is done inplace in source_dir.
    Returns
    -------
    None
    """
    input_files = []

    for file in os.listdir(source_dir):
        if re.match(r'.*(\.bmp|\.jpg|\.png|\.gif)$', file, re.I):
            input_files.append(os.path.join(source_dir, file))

    if not input_files:
        print('No files were found.')
        return

    extension = '.' + input_files[0].split(".")[-1]
    if output_dir is None:
        for i, filename in enumerate(natsorted(input_files)):
            os.rename(filename, os.path.join(source_dir, str(i) + extension))
        print(f'Files within {source_dir} were renamed, starting from 0{extension} to {i}{extension}.')
    else:
        if not os.path.isdir(output_dir):
            os.mkdir(output_dir)

        for i, filename in enumerate(natsorted(input_files)):
            shutil.copy(filename, os.path.join(output_dir, str(i) + extension))

        print(
            f'Files from {source_dir} were renamed and saved to {output_dir}, starting from 0{extension} to {i}{extension}.')



import setuptools
import os        

with open("README.md", "r") as fh:
    long_description = fh.read()

__version__ = "1.1.15"

if os.environ.get('TARGET_ENV'):
    __version__ = __version__ + ".dev" + os.environ['CI_JOB_ID']

setuptools.setup(
    name="improutils",
    author="ImproLab",
    version=__version__,
    author_email="improlab@fit.cvut.cz",
    description="Package with useful functions for BI-SVZ coursework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ImprolabFIT/improutils_package",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "numpy>=1.19.3",
        "Pillow>=8.1.0",
        "matplotlib>=3.4.2",
        "opencv-python>=4.5.3.56",
        "Pylon>=0.4.4",
        "pytesseract>=0.3.7",
        "wheel",
        "natsort>=5.3.3",
        "PyYAML>=5.3.1",
        "qreader>=3.11",
      ],
    python_requires='>=3.8',
)



from .coordconversion import _calc_alfa_metric_factor
from .coordconversion import *

"""
Allows to estimate real world object height based on two points (top and bottom) measured
on image plane.
"""
class HeightEstimator:

    def __init__(self, ref_measurements, vl, vz):
        self._vanish_line = vl
        self._vert_vanish_point = vz
        self._alfa_metric_factor = _calc_alfa_metric_factor(ref_measurements, self._vanish_line,
                                                            self._vert_vanish_point)

    def calc_height(self, top_point, bottom_point):
        """
        Calculates real world height based on top_point and bottom_point measured
        on image plane.

        :param top_point: ndarray
            Top point in reference direction of the object in inhomogeneous format.
        :param bottom_point: ndarray
            Ground plane point of the object in inhomogeneous format.
        :return: float
            Scalar value representing real world height.
        """
        top_point = convert_pt_to_homogenous(top_point)
        bottom_point = convert_pt_to_homogenous(bottom_point)
        # This formula comes from paper Single view metrology by A. Criminisi.
        height = - np.linalg.norm(np.cross(bottom_point, top_point)) / (
                self._alfa_metric_factor * (np.dot(self._vanish_line, bottom_point)) * np.linalg.norm(
            np.cross(self._vert_vanish_point, top_point)))
        return height
